from qgis.core import QgsProject, QgsMessageLog, QgsMapLayer, QgsRasterLayer, QgsSettings
from PyQt5.QtCore import QByteArray
import os


class PAGGFLoader(object):
    """
    Load a version of the PAGGF plugin in function of the project
    which is opened
    """

    def __init__(self, iface):
        self.iface = iface
        self.pjt = QgsProject.instance()
        self.pjt.readProject.connect(
            self.load_new_version_at_pjt_reading)  # when a new project is
        # loaded -> try to load the plugins
        self.pjt.removeAll.connect(
            self.unload_running_version)  # when changing project we unload
        # the plugin
        self.running_version = None  # Version of the plugin which is running

    def initGui(self):
        # At start check if a project is already opened
        self.load_new_version_at_pjt_reading()

    def unload_running_version(self):
        if self.running_version:
            self.running_version.unload()
            # del self.running_version
            self.running_version = None

    def unload(self):
        self.unload_running_version()

    def load_new_version_at_pjt_reading(self):
        """
        When a new project is loaded, load the version of the plugin relative
        to this project
        """
        self.unload_running_version()

        if os.path.split(self.pjt.fileName())[1] == 'paggf_encodage_bm.qgs':
            QgsMessageLog.logMessage('Load BM', 'PAGGF', QgsMessageLog.ALL)
            from .paggf_version_bm import PAGGF_BM
            self.running_version = PAGGF_BM(self.iface)
            self.running_version.initGui()
        elif os.path.split(self.pjt.fileName())[1] == 'paggf_encodage_cid.qgs':
            QgsMessageLog.logMessage('Load CID', 'PAGGF', QgsMessageLog.ALL)
            from .paggf_version_cid import PAGGF_CID
            self.running_version = PAGGF_CID(self.iface)
            self.running_version.initGui()
        elif os.path.split(self.pjt.fileName())[1] == 'paggf_encodage_ef.qgs':
            QgsMessageLog.logMessage('Load EF', 'PAGGF', QgsMessageLog.ALL)
            from .paggf_version_ef import PAGGF_EF
            self.running_version = PAGGF_EF(self.iface)
            self.running_version.initGui()
        else:
            QgsMessageLog.logMessage('Do nothing', 'PAGGF', QgsMessageLog.ALL)
            return

        # Background layer
        s = QgsSettings()
        ortho = s.value("PAGGF/ortho", "")
        ok = False
        if ortho:
            l = QgsRasterLayer(ortho, "Orthophoto")
            ok = l.isValid()
        if not ok:
            l = QgsRasterLayer("crs=EPSG:3857&format&type=xyz&url=http://b.tile.openstreetmap.org/{z}/{x}/{y}.png&zmax=18&zmin=0", "Fond OSM", "wms")
        # add to the end of the layer stack
        l = QgsProject.instance().layerStore().addMapLayer(l)
        QgsProject.instance().layerTreeRoot().addLayer(l)

        # Toolbar positions
        if self.running_version:
            # Restore toolbar positions
            if hasattr(self.running_version, "TOOLBAR_POSITIONS"):
                self.iface.mainWindow().restoreState(QByteArray(self.running_version.TOOLBAR_POSITIONS))

            # Set the active layer
            self.iface.setActiveLayer(self.running_version.default_active_layer())

        # FIXME: waiting for the "trust" option to be merged in qgis
        for ln, lyr in QgsProject.instance().mapLayers().items():
            if lyr.type() == QgsMapLayer.VectorLayer:
                lyr.updateExtents()

